@@ -1,16 +1,18 @@
 //! Linear algebra operations for OLS estimation and robust variance computation.
 //!
 //! This module provides optimized implementations of:
-//! - OLS solving using LAPACK
+//! - OLS solving using pure Rust (faer library)
 //! - HC1 (heteroskedasticity-consistent) variance-covariance estimation
 //! - Cluster-robust variance-covariance estimation
 
 use ndarray::{Array1, Array2, ArrayView1, ArrayView2, Axis};
-use ndarray_linalg::{FactorizeC, Solve, SolveC, SVD, UPLO};
-use numpy::{IntoPyArray, PyArray1, PyArray2, PyReadonlyArray1, PyReadonlyArray2};
+use numpy::{PyArray1, PyArray2, PyReadonlyArray1, PyReadonlyArray2, ToPyArray};
 use pyo3::prelude::*;
 use std::collections::HashMap;
 
+// faer for pure Rust linear algebra (no external BLAS/LAPACK dependencies)
+use faer::linalg::solvers::{PartialPivLu, Solve};
+
 /// Solve OLS regression: β = (X'X)^{-1} X'y
 ///
 /// Uses SVD with truncation for rank-deficient matrices:
@@ -46,14 +48,14 @@ pub fn solve_ols<'py>(
     cluster_ids: Option<PyReadonlyArray1<'py, i64>>,
     return_vcov: bool,
 ) -> PyResult<(
-    &'py PyArray1<f64>,
-    &'py PyArray1<f64>,
-    Option<&'py PyArray2<f64>>,
+    Bound<'py, PyArray1<f64>>,
+    Bound<'py, PyArray1<f64>>,
+    Option<Bound<'py, PyArray2<f64>>>,
 )> {
     let x_arr = x.as_array();
     let y_arr = y.as_array();
 
-    let _n = x_arr.nrows();
+    let n = x_arr.nrows();
     let k = x_arr.ncols();
 
     // Solve using SVD with truncation for rank-deficient matrices
@@ -61,17 +63,49 @@ pub fn solve_ols<'py>(
     let x_owned = x_arr.to_owned();
     let y_owned = y_arr.to_owned();
 
-    // Compute SVD: X = U * S * V^T
-    let (u_opt, s, vt_opt) = x_owned
-        .svd(true, true)
-        .map_err(|e| PyErr::new::<pyo3::exceptions::PyValueError, _>(format!("SVD failed: {}", e)))?;
+    // Convert ndarray to faer for SVD computation
+    let x_faer = ndarray_to_faer(&x_owned);
 
-    let u = u_opt.ok_or_else(|| {
-        PyErr::new::<pyo3::exceptions::PyValueError, _>("SVD did not return U matrix")
-    })?;
-    let vt = vt_opt.ok_or_else(|| {
-        PyErr::new::<pyo3::exceptions::PyValueError, _>("SVD did not return V^T matrix")
-    })?;
+    // Compute thin SVD using faer: X = U * S * V^T
+    let svd = match x_faer.thin_svd() {
+        Ok(s) => s,
+        Err(_) => {
+            return Err(PyErr::new::<pyo3::exceptions::PyValueError, _>(
+                "SVD computation failed"
+            ))
+        }
+    };
+
+    // Extract U, S, V from faer SVD result (capitalized methods in faer 0.24)
+    let u_faer = svd.U();
+    let s_diag = svd.S();  // Returns diagonal view
+    let s_col = s_diag.column_vector();  // Get as column vector
+    let v_faer = svd.V();  // This is V, not V^T
+
+    // Convert back to ndarray
+    let n_rows = u_faer.nrows();
+    let n_svd_cols = u_faer.ncols();
+    let mut u = Array2::<f64>::zeros((n_rows, n_svd_cols));
+    for i in 0..n_rows {
+        for j in 0..n_svd_cols {
+            u[[i, j]] = u_faer[(i, j)];
+        }
+    }
+
+    let s_len = s_col.nrows();
+    let mut s = Array1::<f64>::zeros(s_len);
+    for i in 0..s_len {
+        s[i] = s_col[i];  // S column vector
+    }
+
+    let v_rows = v_faer.nrows();
+    let v_cols = v_faer.ncols();
+    let mut vt = Array2::<f64>::zeros((v_cols, v_rows));  // V^T has shape (k, k)
+    for i in 0..v_rows {
+        for j in 0..v_cols {
+            vt[[j, i]] = v_faer[(i, j)];  // Transpose V to get V^T
+        }
+    }
 
     // Compute rcond threshold to match R's lm() behavior
     // R's qr() uses tol = 1e-07 by default, which is sqrt(eps) ≈ 1.49e-08
@@ -85,9 +119,10 @@ pub fn solve_ols<'py>(
     let uty = u.t().dot(&y_owned); // (min(n,k),)
 
     // Build S^{-1} with truncation and count effective rank
-    let mut s_inv_uty = Array1::<f64>::zeros(k);
+    // Note: s.len() = min(n, k) from thin SVD, so this handles underdetermined (n < k) correctly
+    let mut s_inv_uty = Array1::<f64>::zeros(s.len());
     let mut rank = 0usize;
-    for i in 0..s.len().min(k) {
+    for i in 0..s.len() {
         if s[i] > threshold {
             s_inv_uty[i] = uty[i] / s[i];
             rank += 1;
@@ -109,20 +144,20 @@ pub fn solve_ols<'py>(
             // Rank-deficient: cannot compute valid vcov, return NaN matrix
             let mut nan_vcov = Array2::<f64>::zeros((k, k));
             nan_vcov.fill(f64::NAN);
-            Some(nan_vcov.into_pyarray(py))
+            Some(nan_vcov.to_pyarray_bound(py))
         } else {
             // Full rank: compute robust vcov normally
             let cluster_arr = cluster_ids.as_ref().map(|c| c.as_array().to_owned());
-            let vcov_arr = compute_robust_vcov_internal(&x_arr, &residuals.view(), cluster_arr.as_ref())?;
-            Some(vcov_arr.into_pyarray(py))
+            let vcov_arr = compute_robust_vcov_internal(&x_arr, &residuals.view(), cluster_arr.as_ref(), n, k)?;
+            Some(vcov_arr.to_pyarray_bound(py))
         }
     } else {
         None
     };
 
     Ok((
-        coefficients.into_pyarray(py),
-        residuals.into_pyarray(py),
+        coefficients.to_pyarray_bound(py),
+        residuals.to_pyarray_bound(py),
         vcov,
     ))
 }
@@ -143,13 +178,15 @@ pub fn compute_robust_vcov<'py>(
     x: PyReadonlyArray2<'py, f64>,
     residuals: PyReadonlyArray1<'py, f64>,
     cluster_ids: Option<PyReadonlyArray1<'py, i64>>,
-) -> PyResult<&'py PyArray2<f64>> {
+) -> PyResult<Bound<'py, PyArray2<f64>>> {
     let x_arr = x.as_array();
     let residuals_arr = residuals.as_array();
     let cluster_arr = cluster_ids.as_ref().map(|c| c.as_array().to_owned());
 
-    let vcov = compute_robust_vcov_internal(&x_arr, &residuals_arr, cluster_arr.as_ref())?;
-    Ok(vcov.into_pyarray(py))
+    let n = x_arr.nrows();
+    let k = x_arr.ncols();
+    let vcov = compute_robust_vcov_internal(&x_arr, &residuals_arr, cluster_arr.as_ref(), n, k)?;
+    Ok(vcov.to_pyarray_bound(py))
 }
 
 /// Internal implementation of robust variance-covariance computation.
@@ -157,14 +194,13 @@ fn compute_robust_vcov_internal(
     x: &ArrayView2<f64>,
     residuals: &ArrayView1<f64>,
     cluster_ids: Option<&Array1<i64>>,
+    n: usize,
+    k: usize,
 ) -> PyResult<Array2<f64>> {
-    let n = x.nrows();
-    let k = x.ncols();
-
     // Compute X'X
     let xtx = x.t().dot(x);
 
-    // Compute (X'X)^{-1} using Cholesky decomposition
+    // Compute (X'X)^{-1} using LU decomposition
     let xtx_inv = invert_symmetric(&xtx)?;
 
     match cluster_ids {
@@ -240,58 +276,110 @@ fn compute_robust_vcov_internal(
     }
 }
 
+/// Convert ndarray Array2 to faer Mat
+fn ndarray_to_faer(arr: &Array2<f64>) -> faer::Mat<f64> {
+    let nrows = arr.nrows();
+    let ncols = arr.ncols();
+    faer::Mat::from_fn(nrows, ncols, |i, j| arr[[i, j]])
+}
+
 /// Invert a symmetric positive-definite matrix.
 ///
-/// Tries Cholesky factorization first (faster for well-conditioned SPD matrices),
-/// falls back to LU decomposition for near-singular or indefinite matrices.
+/// Uses LU decomposition with partial pivoting. Includes both NaN/Inf check
+/// and conditional residual-based verification to catch near-singular matrices
+/// that produce finite but numerically inaccurate results.
 ///
-/// Cholesky (when applicable):
-/// - ~2x faster than LU decomposition
-/// - More numerically stable for positive-definite matrices
-/// - Reuses the factorization across all column solves
+/// Performance optimization: The expensive O(n³) residual check (A * A⁻¹ - I)
+/// is only performed when LU pivot ratios suggest potential instability. For
+/// well-conditioned matrices (the common case), this check is skipped.
 fn invert_symmetric(a: &Array2<f64>) -> PyResult<Array2<f64>> {
     let n = a.nrows();
 
-    // Try Cholesky factorization first (faster for well-conditioned SPD matrices)
-    if let Ok(factorized) = a.factorizec(UPLO::Lower) {
-        // Solve A X = I for each column using Cholesky
-        let mut result = Array2::<f64>::zeros((n, n));
-        let mut cholesky_failed = false;
+    // Convert ndarray to faer
+    let a_faer = ndarray_to_faer(a);
 
-        for i in 0..n {
-            let mut e_i = Array1::<f64>::zeros(n);
-            e_i[i] = 1.0;
+    // Create identity matrix in faer
+    let identity = faer::Mat::from_fn(n, n, |i, j| if i == j { 1.0 } else { 0.0 });
 
-            match factorized.solvec(&e_i) {
-                Ok(col) => result.column_mut(i).assign(&col),
-                Err(_) => {
-                    cholesky_failed = true;
-                    break;
-                }
+    // Use LU decomposition with partial pivoting
+    let lu = PartialPivLu::new(a_faer.as_ref());
+
+    // Solve A * X = I  =>  X = A^{-1}
+    let x_faer = lu.solve(&identity);
+
+    // Check for NaN/Inf in result (indicates singular matrix)
+    let mut has_nan = false;
+    for i in 0..n {
+        for j in 0..n {
+            if !x_faer[(i, j)].is_finite() {
+                has_nan = true;
+                break;
             }
         }
-
-        if !cholesky_failed {
-            return Ok(result);
+        if has_nan {
+            break;
         }
     }
 
-    // Fallback to LU decomposition for near-singular or indefinite matrices
+    if has_nan {
+        return Err(PyErr::new::<pyo3::exceptions::PyValueError, _>(
+            "Matrix inversion failed (singular matrix)"
+        ));
+    }
+
+    // Check pivot ratio to detect potential instability.
+    // The diagonal of U contains the pivots from LU factorization.
+    // A small pivot ratio (min/max) indicates potential numerical instability.
+    let u_factor = lu.U();
+    let mut max_pivot = 0.0_f64;
+    let mut min_pivot = f64::INFINITY;
+    for i in 0..n {
+        let pivot = u_factor[(i, i)].abs();
+        if pivot > 0.0 {
+            max_pivot = max_pivot.max(pivot);
+            min_pivot = min_pivot.min(pivot);
+        }
+    }
+    let pivot_ratio = if max_pivot > 0.0 { min_pivot / max_pivot } else { 0.0 };
+
+    // Only perform expensive residual check if pivots suggest potential instability.
+    // Threshold of 1e-10 catches truly problematic matrices while avoiding
+    // unnecessary O(n³) computation for well-conditioned cases.
+    if pivot_ratio < 1e-10 {
+        // Verify inversion accuracy by checking ||A * A^{-1} - I||_max
+        // For near-singular matrices, this residual will be large even if
+        // the result contains no NaN/Inf values
+        let a_times_inv = a_faer.as_ref() * &x_faer;
+        let mut max_residual = 0.0_f64;
+        for i in 0..n {
+            for j in 0..n {
+                let expected = if i == j { 1.0 } else { 0.0 };
+                let residual = (a_times_inv[(i, j)] - expected).abs();
+                max_residual = max_residual.max(residual);
+            }
+        }
+
+        // Threshold: detect truly singular matrices while allowing ill-conditioned ones
+        // Ill-conditioned matrices (high condition number) can have residuals up to ~1e-4
+        // while still producing usable results. Use 1e-4 * n as threshold.
+        let threshold = 1e-4 * (n as f64);
+        if max_residual > threshold {
+            return Err(PyErr::new::<pyo3::exceptions::PyValueError, _>(
+                format!(
+                    "Matrix inversion numerically unstable (residual={:.2e} > threshold={:.2e}). \
+                     Design matrix may be near-singular.",
+                    max_residual, threshold
+                )
+            ));
+        }
+    }
+
+    // Convert back to ndarray
     let mut result = Array2::<f64>::zeros((n, n));
     for i in 0..n {
-        let mut e_i = Array1::<f64>::zeros(n);
-        e_i[i] = 1.0;
-
-        let col = a.solve(&e_i).map_err(|e| {
-            PyErr::new::<pyo3::exceptions::PyValueError, _>(format!(
-                "Matrix inversion failed (likely rank-deficient X'X): {}. \
-                 If the design matrix is rank-deficient, use solve_ols without \
-                 skip_rank_check=True to enable R-style handling.",
-                e
-            ))
-        })?;
-
-        result.column_mut(i).assign(&col);
+        for j in 0..n {
+            result[[i, j]] = x_faer[(i, j)];
+        }
     }
 
     Ok(result)
@@ -314,4 +402,48 @@ mod tests {
         assert!((identity[[0, 1]]).abs() < 1e-10);
         assert!((identity[[1, 0]]).abs() < 1e-10);
     }
+
+    #[test]
+    fn test_ndarray_to_faer() {
+        let arr = array![[1.0, 2.0], [3.0, 4.0]];
+        let faer_mat = ndarray_to_faer(&arr);
+        assert_eq!(faer_mat[(0, 0)], 1.0);
+        assert_eq!(faer_mat[(0, 1)], 2.0);
+        assert_eq!(faer_mat[(1, 0)], 3.0);
+        assert_eq!(faer_mat[(1, 1)], 4.0);
+    }
+
+    #[test]
+    fn test_svd_underdetermined_dimensions() {
+        // Underdetermined system: n=2 observations, k=3 coefficients
+        // X is (2, 3), y is (2,)
+        // This test verifies that thin SVD returns the correct dimensions
+        // for underdetermined systems and that our code handles them correctly
+        let x = array![[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]];
+        let _y = array![7.0, 8.0];
+
+        // Convert to faer and compute thin SVD
+        let x_faer = ndarray_to_faer(&x);
+        let svd = x_faer.thin_svd().unwrap();
+
+        // For n=2 < k=3: U is (2, 2), S has 2 values, V is (3, 2)
+        assert_eq!(svd.U().nrows(), 2, "U should have n=2 rows");
+        assert_eq!(svd.U().ncols(), 2, "U should have min(n,k)=2 cols");
+        assert_eq!(svd.S().column_vector().nrows(), 2, "S should have min(n,k)=2 singular values");
+        assert_eq!(svd.V().nrows(), 3, "V should have k=3 rows");
+        assert_eq!(svd.V().ncols(), 2, "V should have min(n,k)=2 cols");
+
+        // Verify s_inv_uty dimension calculation
+        let s_len = svd.S().column_vector().nrows();
+        assert_eq!(s_len, 2, "s.len() should be min(n,k)=2, not k=3");
+
+        // This is the key fix: s_inv_uty must have dimension s.len()=min(n,k),
+        // not k, otherwise vt.t().dot(&s_inv_uty) will have mismatched dimensions
+    }
+
+    // Note: Singular and near-singular matrix tests removed because:
+    // 1. invert_symmetric() returns PyResult, which requires Python initialization
+    //    to create PyErr - `cargo test` without Python causes panic
+    // 2. These edge cases are tested at the Python integration level in
+    //    tests/test_linalg.py with proper fallback handling
 }
